#FLM: TR: Contour
# -----------------------------------------------------------
# (C) Vassil Kateliev, 2018-2021 	(http://www.kateliev.com)
# (C) Karandash Type Foundry 		(http://www.karandash.eu)
#------------------------------------------------------------

# No warranties. By using this you agree
# that you use it at your own risk!

# - Dependencies -----------------
from __future__ import absolute_import, print_function
from collections import OrderedDict
from itertools import groupby

import fontlab as fl6
import fontgate as fgt

from typerig.core.base.message import *
from typerig.proxy.fl.objects.base import Coord
from typerig.proxy.fl.objects.glyph import eGlyph
from typerig.proxy.fl.objects.shape import eShape
from typerig.proxy.fl.objects.contour import eContour
from typerig.proxy.fl.objects.font import pFontMetrics

from PythonQt import QtCore
from typerig.proxy.fl.gui import QtGui
from typerig.proxy.fl.gui.widgets import getProcessGlyphs, TRTransformCtrl

# - Init -------------------------
global pLayers
global pMode
pLayers = None
pMode = 0
app_name, app_version = 'TypeRig | Contour', '2.2'

# - Sub widgets ------------------------
class breakContour(QtGui.QGridLayout):
	# - Split/Break contour 
	def __init__(self):
		super(breakContour, self).__init__()
			 
		# -- Split button
		self.btn_splitContour = QtGui.QPushButton('&Break')
		self.btn_splitContourClose = QtGui.QPushButton('&Break && Close')
		
		self.btn_splitContour.clicked.connect(self.splitContour)
		self.btn_splitContourClose.clicked.connect(self.splitContourClose)
		
		self.btn_splitContour.setMinimumWidth(80)
		self.btn_splitContourClose.setMinimumWidth(80)

		self.btn_splitContour.setToolTip('Break contour at selected Node(s).')
		self.btn_splitContourClose.setToolTip('Break contour and close open contours!\nUseful for cutting stems and etc.')

		# -- Extrapolate value
		self.edt_expand = QtGui.QLineEdit('0.0')
		#self.edt_expand.setMinimumWidth(30)

		self.edt_expand.setToolTip('Extrapolate endings.')
								
		# -- Build: Split/Break contour
		self.addWidget(self.btn_splitContour, 0, 0)
		self.addWidget(QtGui.QLabel('E:'), 0, 1)
		self.addWidget(self.edt_expand, 0, 2)
		self.addWidget(self.btn_splitContourClose, 0, 3)
				
	def splitContour(self):
		glyph = eGlyph()
		glyph.splitContour(layers=pLayers, expand=float(self.edt_expand.text), close=False)
		glyph.updateObject(glyph.fl, 'Break Contour @ %s.' %'; '.join(glyph._prepareLayers(pLayers)))
		glyph.update()

	def splitContourClose(self):
		glyph = eGlyph()
		glyph.splitContour(layers=pLayers, expand=float(self.edt_expand.text), close=True)
		glyph.updateObject(glyph.fl, 'Break Contour & Close @ %s.' %'; '.join(glyph._prepareLayers(pLayers)))
		glyph.update()        

class basicContour(QtGui.QGridLayout):
	# - Split/Break contour 
	def __init__(self):

		# - Init
		self.var_bool_A = {}
		self.var_bool_B = {}

		# - Widgets
		# -- Buttons
		super(basicContour, self).__init__()
		self.btn_BL = QtGui.QPushButton('B L')
		self.btn_TL = QtGui.QPushButton('T L')
		self.btn_BR = QtGui.QPushButton('B R')
		self.btn_TR = QtGui.QPushButton('T R')
		self.btn_sort_x = QtGui.QPushButton('Sort X')
		self.btn_sort_y = QtGui.QPushButton('Sort Y')
		self.btn_close = QtGui.QPushButton('C&lose')
		self.btn_start = QtGui.QPushButton('&Start')
		self.btn_CW = QtGui.QPushButton('CW')
		self.btn_CCW = QtGui.QPushButton('CCW')
		self.btn_overlap = QtGui.QPushButton('Remove Overlap')

		self.btn_bool_setA = QtGui.QPushButton('A')
		self.btn_bool_setB = QtGui.QPushButton('B')
		self.btn_bool_subtract = QtGui.QPushButton('A - B')

		self.btn_bool_setA.setCheckable(True)
		self.btn_bool_setB.setCheckable(True)
		self.btn_bool_subtract.setEnabled(False)

		self.btn_close.setMinimumWidth(40)
		self.btn_BL.setMinimumWidth(40)
		self.btn_TL.setMinimumWidth(40)
		self.btn_BR.setMinimumWidth(40)
		self.btn_TR.setMinimumWidth(40)
		self.btn_start.setMinimumWidth(40)
		self.btn_CW.setMinimumWidth(40)
		self.btn_CCW.setMinimumWidth(40)

		self.btn_close.setToolTip('Close selected contour')
		self.btn_start.setToolTip('Set start point:\n Selected Node') 
		self.btn_BL.setToolTip('Set start point:\nBottom Left Node') 
		self.btn_TL.setToolTip('Set start point:\nTop Left Node') 
		self.btn_BR.setToolTip('Set start point:\nBottom Right Node') 
		self.btn_TR.setToolTip('Set start point:\nTop Right Node') 
		self.btn_CW.setToolTip('Set direction:\nClockwise (TT)') 
		self.btn_CCW.setToolTip('Set direction:\nCounterclockwise (PS)') 
		self.btn_sort_x.setToolTip('Reorder contours within shapes based on their X coordinate.\nSHIFT+Click reverses order!')
		self.btn_sort_y.setToolTip('Reorder contours within shapes based on their Y coordinate.\nSHIFT+Click reverses order!')
		
		self.btn_start.clicked.connect(lambda : self.setStartSelection())
		self.btn_BL.clicked.connect(lambda : self.setStart((0,0)))
		self.btn_TL.clicked.connect(lambda : self.setStart((0,1)))
		self.btn_BR.clicked.connect(lambda : self.setStart((1,0)))
		self.btn_TR.clicked.connect(lambda : self.setStart((1,1)))
		self.btn_CW.clicked.connect(lambda : self.setDirection(False))
		self.btn_CCW.clicked.connect(lambda : self.setDirection(True))
		self.btn_close.clicked.connect(self.closeContour)
		self.btn_overlap.clicked.connect(self.removeOverlap)
		self.btn_sort_x.clicked.connect(lambda : self.setOrder(False))
		self.btn_sort_y.clicked.connect(lambda : self.setOrder(True))
		self.btn_bool_setA.clicked.connect(lambda: self.bool_set_shape(False))
		self.btn_bool_setB.clicked.connect(lambda: self.bool_set_shape(True))
		self.btn_bool_subtract.clicked.connect(lambda: self.bool_subtract())

		self.addWidget(self.btn_close, 						0, 0, 1, 1)
		self.addWidget(self.btn_start, 						0, 1, 1, 1)
		self.addWidget(self.btn_CW, 						1, 0, 1, 1)
		self.addWidget(self.btn_CCW, 						1, 1, 1, 1)

		self.addWidget(self.btn_TL, 						0, 2, 1, 1)
		self.addWidget(self.btn_TR, 						0, 3, 1, 1)
		self.addWidget(self.btn_BL, 						1, 2, 1, 1)
		self.addWidget(self.btn_BR, 						1, 3, 1, 1)
		self.addWidget(self.btn_sort_x, 					2, 0, 1, 2)
		self.addWidget(self.btn_sort_y, 					2, 2, 1, 2)

		self.addWidget(QtGui.QLabel('Boolean Operations:'), 3, 0, 1, 4)
		self.addWidget(self.btn_overlap, 					4, 0, 1, 4)
		self.addWidget(self.btn_bool_setA, 					5, 0, 1, 1)
		self.addWidget(self.btn_bool_setB, 					5, 1, 1, 1)
		self.addWidget(self.btn_bool_subtract, 				5, 2, 1, 2)
		
	def closeContour(self):
		glyph = eGlyph()
		wLayers = glyph._prepareLayers(pLayers)
		selection = glyph.selectedAtContours()

		for layerName in wLayers:
			contours = glyph.contours(layerName)

			for cID, nID in reversed(selection):
				if not contours[cID].closed: contours[cID].closed = True

		glyph.updateObject(glyph.fl, 'Close Contour @ %s.' %'; '.join(wLayers))
		glyph.update()

	def bool_set_shape(self, isB=False):
		# - Init
		glyph = eGlyph()
		selection = glyph.selectedAtShapes()
		work_selection = {}
		reset_value = False

		# - Prepare
		for sid, cid, nid in selection:
			work_selection.setdefault(sid, set([])).add(cid)

		# - Set
		if self.btn_bool_setA.isChecked() and not isB:
			self.var_bool_A = work_selection
		
		elif not self.btn_bool_setA.isChecked() and not isB:
			self.var_bool_A = {}
			reset_value = True
		
		if self.btn_bool_setB.isChecked() and isB:
			self.var_bool_B = work_selection

		elif not self.btn_bool_setB.isChecked() and isB:
			self.var_bool_B = {}
			reset_value = True

		if not reset_value:
			output(0, app_name, 'Boolean Operations | Set {} for contours: {}'.format(['A', 'B'][isB], work_selection))
		else:
			output(0, app_name, 'Boolean Operations | Reset {}'.format(['A', 'B'][isB]))

	def bool_subtract(self):
		glyph = eGlyph()
		wLayers = glyph._prepareLayers(pLayers)

		# - Get contours
		process_shapes = []

		if len(self.var_bool_A.keys()) and len(self.var_bool_B.keys()):
			for layerName in wLayers:
				process_shapes_A = []
				process_shapes_B = []

				for sid, cid_list in self.var_bool_A.items():
					process_shapes_A.append((glyph.shapes(layerName)[sid], [glyph.contours(layerName)[cid] for cid in cid_list]))

				for sid, cid_list in self.var_bool_B.items():
					process_shapes_B.append((glyph.shapes(layerName)[sid], [glyph.contours(layerName)[cid] for cid in cid_list]))

				process_shapes.append(zip(process_shapes_A, process_shapes_B))

		for source_A, source_B in sum(process_shapes, []):
			# - A
			shape_A, contours_A = source_A
			shape_A.removeContours(contours_A)
			new_shape_A = fl6.flShape()
			new_fg_shape_A = fgt.fgShape()
			new_shape_A.contours = contours_A
			new_shape_A.convertToFgShape(new_fg_shape_A)

			# - B
			shape_B, contours_B = source_B
			shape_B.removeContours(contours_B)
			new_shape_B = fl6.flShape()
			new_fg_shape_B = fgt.fgShape()
			new_shape_B.contours = contours_B
			new_shape_B.convertToFgShape(new_fg_shape_B)

			print(new_fg_shape_A.contours, new_fg_shape_B.contours)

			# - Boolean
			result = new_fg_shape_A.subtractShape(new_fg_shape_B, 0)
			shape_A.addContours(fl6.flShape(new_fg_shape_A).contours, True)
			shape_B.addContours(fl6.flShape(new_fg_shape_B).contours, True)

		glyph.updateObject(glyph.fl, 'Boolean Operations @ %s.' %'; '.join(wLayers))
		glyph.update()

	def removeOverlap(self):
		glyph = eGlyph()
		wLayers = glyph._prepareLayers(pLayers)
		
		# - Prepare selection
		tmp = {}
		selection = glyph.selectedAtShapes()

		for sid, cid, nid in selection:
			tmp.setdefault(sid,[]).append(cid)

		selection = {key:list(set(value)) for key, value in tmp.items()}

		# - Get contours
		process_shapes = []
		
		for layerName in wLayers:
			for sid, cid_list in selection.items():
				process_shapes.append((glyph.shapes(layerName)[sid], [glyph.contours(layerName)[cid] for cid in cid_list]))

		for shape, contours in process_shapes:
			shape.removeContours(contours)
			new_shape = fl6.flShape()
			new_fg_shape = fgt.fgShape()
			new_shape.contours = contours
			new_shape.convertToFgShape(new_fg_shape)
			new_fg_shape.removeOverlap()
			shape.addContours(fl6.flShape(new_fg_shape).contours, True)

		glyph.updateObject(glyph.fl, 'Remove Overlap @ %s.' %'; '.join(wLayers))
		glyph.update()

	def setStartSelection(self):
		process_glyphs = getProcessGlyphs(pMode)

		for glyph in process_glyphs:
			wLayers = glyph._prepareLayers(pLayers, False)
			
			selected_contours = {layer:glyph.selectedAtContours(layer)[0] for layer in wLayers}

			for layer, selection in selected_contours.items():
				cid, nid = selection
				glyph.contours(layer)[cid].setStartPoint(nid)

			glyph.update()
			glyph.updateObject(glyph.fl, 'Glyph: %s;\tManual: Set Start Point @ %s.' %(glyph.name, '; '.join(wLayers)))

	def setOrder(self, sort_y=False):
		process_glyphs = getProcessGlyphs(pMode)
		modifiers = QtGui.QApplication.keyboardModifiers()

		for glyph in process_glyphs:
			wLayers = glyph._prepareLayers(pLayers, False)

			for layerName in wLayers:
				wShapes = glyph.shapes(layerName, extend=eShape)

				for shape in wShapes:
					if sort_y:
						shape.contourOrder((None, not modifiers == QtCore.Qt.ShiftModifier)) # Shift reverses
					else:
						shape.contourOrder((True, not modifiers == QtCore.Qt.ShiftModifier))

			glyph.update()
			glyph.updateObject(glyph.fl, 'Glyph: %s;\tAction: Reorder contours @ %s.' %(glyph.name, '; '.join(wLayers)))

	def setStart(self, control=(0,0)):
		process_glyphs = getProcessGlyphs(pMode)

		for glyph in process_glyphs:
			wLayers = glyph._prepareLayers(pLayers, False)

			if control == (0,0): 	# BL
				criteria = lambda node : (node.y, node.x)
			elif control == (0,1): 	# TL
				criteria = lambda node : (-node.y, node.x)
			elif control == (1,0): 	# BR
				criteria = lambda node : (node.y, -node.x)
			elif control == (1,1): 	# TR
				criteria = lambda node : (-node.y, -node.x)
			
			for layerName in wLayers:
				contours = glyph.contours(layerName)

				for contour in contours:
					onNodes = [node for node in contour.nodes() if node.isOn()]
					newFirstNode = sorted(onNodes, key=criteria)[0]
					contour.setStartPoint(newFirstNode.index)

			glyph.update()
			glyph.updateObject(glyph.fl, 'Glyph: %s;\tAction: Set Start Points @ %s.' %(glyph.name, '; '.join(wLayers)))

	def setDirection(self, ccw=True):
		process_glyphs = getProcessGlyphs(pMode)

		for glyph in process_glyphs:
			selection = glyph.selectedAtContours()

			wLayers = glyph._prepareLayers(pLayers, False)

			for layerName in wLayers:
				all_contours = glyph.contours(layerName)

				if len(selection):
					process_contours = [eContour(all_contours[item[0]]) for item in selection]
				else:
					process_contours = [eContour(contour) for contour in all_contours]

				for contour in process_contours:
					if ccw:
						contour.setCCW()
					else:
						contour.setCW()

			glyph.update()
			glyph.updateObject(glyph.fl, 'Glyph: %s;\tAction: Set contour direction @ %s.' %(glyph.name, '; '.join(wLayers)))

class alignContours(QtGui.QGridLayout):
	# - Align Contours
	def __init__(self):
		super(alignContours, self).__init__()
				
		# - Init
		self.align_x = OrderedDict([('Left','L'), ('Right','R'), ('Center','C'), ('Keep','K')])
		self.align_y = OrderedDict([('Top','T'), ('Bottom','B'), ('Center','E'), ('Keep','X')])
		self.align_mode = OrderedDict([	('Layer','CL'), 
										('Base to X-Height','CMX'),
										('Base to Caps','CMC'),
										('Base Ascender','CMA'),
										('Descender to Base','CMD'),
										('Contour to Contour','CC'),
										('Contour to Contour (REV)','RC'),
										('Contour to Node','CN'),
										('Contour to Node (REV)','RN')
									])
		
		# !!! To be implemented
		#self.align_mode = OrderedDict([('Layer','CL'), ('Contour to Contour','CC'), ('Contour to Contour (REV)','RC'), ('Contour to Node','CN'),('Node to Node','NN')])

		# - Widgets
		self.cmb_align_x = QtGui.QComboBox()
		self.cmb_align_y = QtGui.QComboBox()
		self.cmb_align_mode = QtGui.QComboBox()
		self.cmb_align_x.addItems(list(self.align_x.keys()))
		self.cmb_align_y.addItems(list(self.align_y.keys()))
		self.cmb_align_mode.addItems(list(self.align_mode.keys()))

		self.cmb_align_x.setToolTip('Horizontal Alignment')
		self.cmb_align_y.setToolTip('Vertical Alignment')
		self.cmb_align_mode.setToolTip('Alignment Mode')

		self.btn_align = QtGui.QPushButton('Align')
		self.btn_align.clicked.connect(self.alignContours)

		self.addWidget(self.cmb_align_mode, 	0, 0, 1, 2)
		self.addWidget(self.cmb_align_x, 		0, 2, 1, 1)
		self.addWidget(self.cmb_align_y, 		0, 3, 1, 1)
		self.addWidget(self.btn_align, 			1, 0, 1, 4)

	def alignContours(self):
		# - Helpers
		def getContourBonds(work_contours):
			tmp_bounds = [contour.bounds for contour in work_contours]
			cont_min_X, cont_min_Y, cont_max_X, cont_max_Y = map(set, zip(*tmp_bounds))
			return (min(cont_min_X), min(cont_min_Y), max(cont_max_X), max(cont_max_Y))

		def getAlignDict(bounds_tuple):
			align_dict = {	'L': bounds_tuple[0], 
							'R': bounds_tuple[2],
							'C': bounds_tuple[2]/2,
							'B': bounds_tuple[1], 
							'T': bounds_tuple[3], 
							'E': bounds_tuple[3]/2
						}

			return align_dict

		# - Init
		user_mode =  self.align_mode[self.cmb_align_mode.currentText]
		user_x = self.align_x[self.cmb_align_x.currentText]
		user_y = self.align_y[self.cmb_align_y.currentText]
		keep_x, keep_y = True, True	

		if user_x == 'K': keep_x = False; user_x = 'L'
		if user_y == 'X': keep_y = False; user_y = 'B'		
		
		process_glyphs = getProcessGlyphs(pMode)

		# - Process
		for glyph in process_glyphs:
			selection = glyph.selectedAtContours()
			wLayers = glyph._prepareLayers(pLayers)

			for layerName in wLayers:
				glyph_contours = glyph.contours(layerName, extend=eContour)
				work_contours = [glyph_contours[index] for index in list(set([item[0] for item in selection]))]
				
				if user_mode =='CC': # Align contours to contours
					if 1 < len(work_contours) < 3:
						c1, c2 = work_contours
						c1.alignTo(c2, user_x + user_y, (keep_x, keep_y))

					elif len(work_contours) > 2:
						cont_bounds = getContourBonds(work_contours)
						align_type = getAlignDict(cont_bounds)
						target = Coord(align_type[user_x], align_type[user_y])

						for contour in work_contours:
							contour.alignTo(target, user_x + user_y, (keep_x, keep_y))
					
				elif user_mode == 'RC': # Align contours to contours in reverse order
					if 1 < len(work_contours) < 3:
						c1, c2 = work_contours
						c2.alignTo(c1, user_x + user_y, (keep_x, keep_y))

					elif len(work_contours) > 2:
						cont_bounds = getContourBonds(work_contours)
						align_type = getAlignDict(cont_bounds)
						target = Coord(align_type[user_x], align_type[user_y])

						for contour in reversed(work_contours):
							contour.alignTo(target, user_x + user_y, (keep_x, keep_y))	

				# !!! To be implemented
				elif user_mode == 'CN': # Align contour to node
					target = work_contours.pop(0)
					target_node_index = selection[0][1]
					target_node = target.fl.nodes()[target_node_index]

					for contour in work_contours:
							contour.alignTo(target_node, user_x + user_y, (keep_x, keep_y))

				elif user_mode == 'RN': # Align contour to node reversed
					target = work_contours.pop()
					target_node_index = selection[-1][1]
					target_node = target.fl.nodes()[target_node_index]

					for contour in work_contours:
							contour.alignTo(target_node, user_x + user_y, (keep_x, keep_y))

				elif user_mode == 'NN': # Align a node on contour to node on another
					pass

				else:
					metrics = pFontMetrics(glyph.package)
					max_layer_y = max([metrics.getXHeight(layerName), metrics.getCapsHeight(layerName), metrics.getAscender(layerName)])
					min_layer_y = min([0, metrics.getDescender(layerName)])
					layer_bounds = QtCore.QRect(0, 0, glyph.getAdvance(layerName), abs(max_layer_y) + abs(min_layer_y))

					if user_mode == 'CL': # Align all contours in given Layer
						cont_bounds = (layer_bounds.x(), layer_bounds.y(), layer_bounds.x() + layer_bounds.width(), layer_bounds.y() + layer_bounds.height())
					
					elif user_mode == 'CMX': # Align all contours to X height
						height = metrics.getXHeight(layerName)
						cont_bounds = (layer_bounds.x(), 0., layer_bounds.x() + layer_bounds.width(), height)

					elif user_mode == 'CMC': # Align all contours to Caps height
						height = metrics.getCapsHeight(layerName)
						cont_bounds = (layer_bounds.x(), 0., layer_bounds.x() + layer_bounds.width(), height)

					elif user_mode == 'CMA': # Align all contours to Ascender height
						height = metrics.getAscender(layerName)
						cont_bounds = (layer_bounds.x(), 0., layer_bounds.x() + layer_bounds.width(), height)

					elif user_mode == 'CMD': # Align all contours to Ascender height
						height = metrics.getDescender(layerName)
						cont_bounds = (layer_bounds.x(), 0., layer_bounds.x() + layer_bounds.width(), height)

					align_type = getAlignDict(cont_bounds)
					target = Coord(align_type[user_x], align_type[user_y])					

					for contour in work_contours:
						contour.alignTo(target, user_x + user_y, (keep_x, keep_y))

			glyph.update()
			glyph.updateObject(glyph.fl, 'Glyph: %s;\tAction: Align Contours @ %s.' %(glyph.name, '; '.join(wLayers)))

class copyContours(QtGui.QGridLayout):
	# - Align Contours
	def __init__(self):
		super(copyContours, self).__init__()

		# - Init
		self.contourClipboard = []

		# -- Custom controls
		self.tr_trans_ctrl = TRTransformCtrl()

		# -- Listview
		self.lst_contours = QtGui.QListView()
		self.mod_contours = QtGui.QStandardItemModel(self.lst_contours)
		self.lst_contours.setModel(self.mod_contours)
		self.lst_contours.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)

		# -- Quick Tool buttons
		self.btn_copy_contour = QtGui.QPushButton('Copy')
		self.btn_paste_contour = QtGui.QPushButton('Paste')
		self.btn_clear = QtGui.QPushButton('Clear')

		self.btn_copy_contour.clicked.connect(self.copy_contour)
		self.btn_paste_contour.clicked.connect(self.paste_contour)
		self.btn_clear.clicked.connect(self.__reset)
				
		self.addWidget(self.btn_copy_contour,		0, 0, 1, 4)
		self.addWidget(self.btn_paste_contour,		0, 4, 1, 4)
		self.addWidget(self.lst_contours,			1, 0, 10, 8)
		self.addWidget(self.btn_clear,				11, 0, 1, 8)

		self.tr_trans_ctrl.lay_controls.setMargin(0)

	def __reset(self):
		self.contourClipboard = []
		self.mod_contours.removeRows(0, self.mod_contours.rowCount())

	def __drawIcon(self, contours, foreground='black', background='white'):
		# - Init
		# -- Prepare contours
		cloned_contours = [contour.clone() for contour in contours]
		new_shape = fl6.flShape()
		for contour in cloned_contours:
			new_shape.addContour(contour, True)


		new_painter = QtGui.QPainter()
		new_icon = QtGui.QIcon()
		shape_bbox = new_shape.boundingBox
		draw_dimension = max(shape_bbox.width(), shape_bbox.height())
		new_pixmap = QtGui.QPixmap(draw_dimension, draw_dimension)

		# - Paint
		new_painter.begin(new_pixmap)
		
		# -- Background
		new_painter.setBrush(QtGui.QBrush(QtGui.QColor(background)))
		new_painter.drawRect(QtCore.QRectF(0, 0, draw_dimension, draw_dimension))
		
		# -- Foreground
		new_painter.setBrush(QtGui.QBrush(QtGui.QColor(foreground)))
		new_transform = new_shape.transform.translate(-shape_bbox.x(), -shape_bbox.y())
		new_shape.applyTransform(new_transform)
		new_shape.ensurePaths()
		new_painter.drawPath(new_shape.closedPath)
		new_icon.addPixmap(new_pixmap.transformed(QtGui.QTransform().scale(1, -1)))
		return new_icon

	def copy_contour(self):
		# - Init
		wGlyph = eGlyph()
		wContours = wGlyph.contours()
		wLayers = wGlyph._prepareLayers(pLayers)
		export_clipboard = OrderedDict()
		
		# - Build initial contour information
		selectionTuples = wGlyph.selectedAtContours()
		selection = {key:[layer_name[1] for layer_name in value] if not wContours[key].isAllNodesSelected() else [] for key, value in groupby(selectionTuples, lambda x:x[0])}

		# - Process
		if len(selection.keys()):
			# -- Add to gallery
			draw_contours = [wGlyph.contours()[cid] for cid in selection.keys()]
			draw_count = sum([contour.nodesCount for contour in draw_contours])
			new_item = QtGui.QStandardItem('{} Nodes | Source: {}'.format(draw_count, wGlyph.name))
			
			if len(wLayers) == len(wGlyph.masters()):
				conditional_color = 'green'
			elif len(wLayers) == 1:
				conditional_color = 'red'
			else:
				conditional_color = 'blue'

			new_icon = self.__drawIcon(draw_contours, conditional_color)
			new_item.setIcon(new_icon)
			new_item.setSelectable(True)
			self.mod_contours.appendRow(new_item)

			# -- Add to clipboard
			for layer_name in wLayers:
				wLayer = layer_name
				export_clipboard[wLayer] = []

				for cid, nList in selection.items():
					if len(nList):
						 export_clipboard[wLayer].append(fl6.flContour([wGlyph.nodes(wLayer)[nid].clone() for nid in nList]))
					else:
						export_clipboard[wLayer].append(wGlyph.contours(wLayer)[cid].clone())

			self.contourClipboard.append(export_clipboard)
			output(0, app_name, 'Copy contours; Glyph: %s; Layers: %s.' %(wGlyph.name, '; '.join(wLayers)))
		
	def paste_contour(self):
		# - Init
		wGlyph = eGlyph()
		wContours = wGlyph.contours()
		wLayers = wGlyph._prepareLayers(pLayers)
		gallery_selection = [qidx.row() for qidx in self.lst_contours.selectedIndexes()]

		# - Helper
		def add_new_shape(layer, contours):
			newShape = fl6.flShape()
			newShape.addContours(contours, True)
			layer.addShape(newShape)

		# - Process
		if len(self.contourClipboard) and len(gallery_selection):
			for idx in gallery_selection:
				paste_data = self.contourClipboard[idx]

				for layerName, contours in paste_data.items():
					wLayer = wGlyph.layer(layerName)
					
					if wLayer is not None:	
						# - Process transform
						cloned_contours = [contour.clone() for contour in contours]

						# - Insert contours into currently selected shape
						try:
							selected_shape = wGlyph.shapes(layerName)[0]
						
						except IndexError:
							selected_shape = fl6.flShape()
							wLayer.addShape(selected_shape)

						selected_shape.addContours(cloned_contours, True)
					
		
			wGlyph.updateObject(wGlyph.fl, 'Paste contours; Glyph: %s; Layers: %s' %(wGlyph.name, '; '.join(wLayers)))

class transformContours(QtGui.QGridLayout):
	# - Align Contours
	def __init__(self):
		super(transformContours, self).__init__()

		# - Init
		# -- Custom controls
		self.tr_trans_ctrl = TRTransformCtrl()

		# -- Quick Tool buttons
		self.btn_transform_contour = QtGui.QPushButton('Transform')
		self.btn_transform_contour.clicked.connect(self.transform_contour)
				
		self.addWidget(self.tr_trans_ctrl, 			0, 0, 2, 8)
		self.addWidget(self.btn_transform_contour,	3, 0, 1, 8)
		
		self.tr_trans_ctrl.lay_controls.setMargin(0)

	def transform_contour(self):
		# - Init
		wGlyph = eGlyph()
		wContours = wGlyph.contours()
		wLayers = wGlyph._prepareLayers(pLayers)
		export_clipboard = OrderedDict()
		
		# - Build initial contour information
		selectionTuples = wGlyph.selectedAtContours()
		selection = [cid for cid, nodes in groupby(selectionTuples, lambda x:x[0])]

		# - Process
		if len(selection):
			for layer_name in wLayers:
				# - Init
				wContours = wGlyph.contours(layer_name)

				# - Get Bounding box
				temp_shape = fl6.flShape()
				temp_contours = [wContours[cid].clone() for cid in selection]
				temp_shape.addContours(temp_contours, True)

				wBBox = temp_shape.boundingBox

				# - Set transformation
				new_transform, org_transform, rev_transform = self.tr_trans_ctrl.getTransform(wBBox)
				
				# - Transform contours
				for cid in selection:
					wContour = wContours[cid]
					wContour.transform = org_transform
					wContour.applyTransform()
					wContour.transform = new_transform
					wContour.applyTransform()
					wContour.transform = rev_transform
					wContour.applyTransform()
					wContour.update()
			
			# - Done
			wGlyph.updateObject(wGlyph.fl, 'Transform contours; Glyph: %s; Layers: %s' %(wGlyph.name, '; '.join(wLayers)))

# - Tabs -------------------------------
class tool_tab(QtGui.QWidget):
	def __init__(self):
		super(tool_tab, self).__init__()

		# - Init
		layoutV = QtGui.QVBoxLayout()
		
		layoutV.addWidget(QtGui.QLabel('Basic Operations:'))
		layoutV.addLayout(basicContour())

		#layoutV.addWidget(QtGui.QLabel('Break/Knot Contour'))
		#layoutV.addLayout(breakContour())

		layoutV.addWidget(QtGui.QLabel('Align:'))
		layoutV.addLayout(alignContours())
		layoutV.addWidget(QtGui.QLabel('Transform:'))
		layoutV.addLayout(transformContours())
		layoutV.addWidget(QtGui.QLabel('Clipboard:'))
		layoutV.addLayout(copyContours())

		# - Build ---------------------------
		layoutV.addStretch()
		self.setLayout(layoutV)

		# !!! Hotfix FL7 7355 
		self.setMinimumSize(300,self.sizeHint.height())


# - Test ----------------------
if __name__ == '__main__':
	test = tool_tab()
	test.setWindowTitle('%s %s' %(app_name, app_version))
	test.setGeometry(100, 100, 200, 400)
	test.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint) # Always on top!!
	
	test.show()